@@ -4,7 +4,7 @@
  * Copyright (C) 2010-2016 Andy Green <andy@warmcat.com>
  *
  * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public
+ * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation:
  * version 2.1 of the License.
  *
@@ -474,7 +474,7 @@ callback_generic_sessions(struct lws *wsi, enum lws_callback_reasons reason,
 					     WSI_TOKEN_HOST) < 0)
 					return 1;
 				lws_snprintf(pss->onward, sizeof(pss->onward) - 1,
-					 "%s%s%s", oprot[lws_is_ssl(wsi)],
+					 "%s%s%s", oprot[!!lws_is_ssl(wsi)],
 					    cookie, args->p);
 				lwsl_notice("redirecting to ourselves with cookie refresh\n");
 				/* we need a redirect to ourselves, session cookie is expired */
